@@ -28,6 +28,9 @@
 #include <limits.h>
 #include <errno.h>
 #include <pthread.h>
+#ifdef HAVE_PTHREAD_NP_H
+# include <pthread_np.h>
+#endif
 #include <time.h>
 #include <sys/param.h>
 #include <sys/queue.h>
@@ -1335,7 +1338,11 @@ httpd_init(void)
 
   exitev = event_new(evbase_httpd, exit_efd, EV_READ, exit_cb, NULL);
 #else
+# ifdef HAVE_PIPE2
   ret = pipe2(exit_pipe, O_CLOEXEC);
+# else
+  ret = pipe(exit_pipe);
+# endif
   if (ret < 0)
     {
       DPRINTF(E_FATAL, L_HTTPD, "Could not create pipe: %s\n", strerror(errno));
@@ -1369,7 +1376,7 @@ httpd_init(void)
       ret = evhttp_bind_socket(evhttpd, "::", port);
       if (ret < 0)
 	{
-	  DPRINTF(E_LOG, L_HTTPD, "Could not bind to port %d, falling back to IPv4\n", port);
+	  DPRINTF(E_LOG, L_HTTPD, "Could not bind to port %d with IPv6, falling back to IPv4\n", port);
 	  v6enabled = 0;
 	}
     }
@@ -1394,6 +1401,12 @@ httpd_init(void)
       goto thread_fail;
     }
 
+#if defined(HAVE_PTHREAD_SETNAME_NP)
+  pthread_setname_np(tid_httpd, "httpd");
+#elif defined(HAVE_PTHREAD_SET_NAME_NP)
+  pthread_set_name_np(tid_httpd, "httpd");
+#endif
+
   return 0;
 
  thread_fail:
